@@ -47,7 +47,7 @@ int Query(char *dev)
 	char *activity;
 
 	if (fd < 0){
-		fprintf(stderr, Name ": cannot open %s: %s\n",
+		pr_err("cannot open %s: %s\n",
 			dev, strerror(errno));
 		return 1;
 	}
@@ -76,7 +76,7 @@ int Query(char *dev)
 	else {
 		printf("%s: %s %s %d devices, %d spare%s. Use mdadm --detail for more detail.\n",
 		       dev,
-		       human_size_brief(larray_size),
+		       human_size_brief(larray_size,IEC),
 		       map_num(pers, array.level),
 		       array.raid_disks,
 		       array.spare_disks, array.spare_disks==1?"":"s");
@@ -124,4 +124,3 @@ int Query(char *dev)
 	}
 	return 0;
 }
-
