@@ -20,8 +20,6 @@
 
 #include "mdadm.h"
 
-#define min(a,b) (((a) < (b)) ? (a) : (b))
-
 inline void sb_le_to_cpu(bitmap_super_t *sb)
 {
 	sb->magic = __le32_to_cpu(sb->magic);
@@ -99,7 +97,7 @@ int count_dirty_bits(char *buf, int num_bits)
 {
 	int i, num = 0;
 
-	for (i=0; i < num_bits / 8; i++)
+	for (i = 0; i < num_bits / 8; i++)
 		num += count_dirty_bits_byte(buf[i], 8);
 
 	if (num_bits % 8) /* not an even byte boundary */
@@ -115,6 +113,14 @@ unsigned long long bitmap_bits(unsigned long long array_size,
 	return (array_size * 512 + chunksize - 1) / chunksize;
 }
 
+unsigned long bitmap_sectors(struct bitmap_super_s *bsb)
+{
+	unsigned long long bits = bitmap_bits(__le64_to_cpu(bsb->sync_size),
+					      __le32_to_cpu(bsb->chunksize));
+	int bits_per_sector = 8*512;
+	return (bits + bits_per_sector - 1) / bits_per_sector;
+}
+
 bitmap_info_t *bitmap_fd_read(int fd, int brief)
 {
 	/* Note: fd might be open O_DIRECT, so we must be
@@ -122,38 +128,30 @@ bitmap_info_t *bitmap_fd_read(int fd, int brief)
 	 */
 	unsigned long long total_bits = 0, read_bits = 0, dirty_bits = 0;
 	bitmap_info_t *info;
-	char *buf, *unaligned;
-	int n, skip;
+	void *buf;
+	unsigned int n, skip;
 
-	unaligned = malloc(8192*2);
-	buf = (char*) ((unsigned long)unaligned | 8191)+1;
-	n = read(fd, buf, 8192);
-
-	info = malloc(sizeof(*info));
-	if (info == NULL) {
-#if __GNUC__ < 3
-		fprintf(stderr, Name ": failed to allocate %d bytes\n",
-				(int)sizeof(*info));
-#else
-		fprintf(stderr, Name ": failed to allocate %zd bytes\n",
-				sizeof(*info));
-#endif
+	if (posix_memalign(&buf, 4096, 8192) != 0) {
+		pr_err("failed to allocate 8192 bytes\n");
 		return NULL;
 	}
+	n = read(fd, buf, 8192);
+
+	info = xmalloc(sizeof(*info));
 
 	if (n < sizeof(info->sb)) {
-		fprintf(stderr, Name ": failed to read superblock of bitmap "
-			"file: %s\n", strerror(errno));
+		pr_err("failed to read superblock of bitmap "
+		       "file: %s\n", strerror(errno));
 		free(info);
-		free(unaligned);
+		free(buf);
 		return NULL;
 	}
 	memcpy(&info->sb, buf, sizeof(info->sb));
 	skip = sizeof(info->sb);
 
 	sb_le_to_cpu(&info->sb); /* convert superblock to CPU byte ordering */
-	
-	if (brief || info->sb.sync_size == 0)
+
+	if (brief || info->sb.sync_size == 0 || info->sb.chunksize == 0)
 		goto out;
 
 	/* read the rest of the file counting total bits and dirty bits --
@@ -184,12 +182,13 @@ bitmap_info_t *bitmap_fd_read(int fd, int brief)
 	}
 
 	if (read_bits < total_bits) { /* file truncated... */
-		fprintf(stderr, Name ": WARNING: bitmap file is not large "
+		pr_err("WARNING: bitmap file is not large "
 			"enough for array size %llu!\n\n",
 			(unsigned long long)info->sb.sync_size);
 		total_bits = read_bits;
 	}
 out:
+	free(buf);
 	info->total_bits = total_bits;
 	info->dirty_bits = dirty_bits;
 	return info;
@@ -203,14 +202,14 @@ bitmap_info_t *bitmap_file_read(char *filename, int brief, struct supertype **st
 	struct supertype *st = *stp;
 
 	if (stat(filename, &stb) < 0) {
-		fprintf(stderr, Name ": failed to find file %s: %s\n",
+		pr_err("failed to find file %s: %s\n",
 			filename, strerror(errno));
 		return NULL;
 	}
 	if ((S_IFMT & stb.st_mode) == S_IFBLK) {
-		fd = open(filename, O_RDONLY);
+		fd = open(filename, O_RDONLY|O_DIRECT);
 		if (fd < 0) {
-			fprintf(stderr, Name ": failed to open bitmap file %s: %s\n",
+			pr_err("failed to open bitmap file %s: %s\n",
 				filename, strerror(errno));
 			return NULL;
 		}
@@ -219,15 +218,18 @@ bitmap_info_t *bitmap_file_read(char *filename, int brief, struct supertype **st
 		if (!st) {
 			/* just look at device... */
 			lseek(fd, 0, 0);
-		} else {	
-			st->ss->locate_bitmap(st, fd, NULL);
-		}
-		ioctl(fd, BLKFLSBUF, 0); /* make sure we read current data */
+		} else if (!st->ss->locate_bitmap) {
+			pr_err("No bitmap possible with %s metadata\n",
+				st->ss->name);
+			return NULL;
+		} else
+			st->ss->locate_bitmap(st, fd);
+
 		*stp = st;
 	} else {
 		fd = open(filename, O_RDONLY|O_DIRECT);
 		if (fd < 0) {
-			fprintf(stderr, Name ": failed to open bitmap file %s: %s\n",
+			pr_err("failed to open bitmap file %s: %s\n",
 				filename, strerror(errno));
 			return NULL;
 		}
@@ -260,6 +262,8 @@ int ExamineBitmap(char *filename, int brief, struct supertype *st)
 	bitmap_info_t *info;
 	int rv = 1;
 	char buf[64];
+	int swap;
+	__u32 uuid32[4];
 
 	info = bitmap_file_read(filename, brief, &st);
 	if (!info)
@@ -269,29 +273,38 @@ int ExamineBitmap(char *filename, int brief, struct supertype *st)
 	printf("        Filename : %s\n", filename);
 	printf("           Magic : %08x\n", sb->magic);
 	if (sb->magic != BITMAP_MAGIC) {
-		fprintf(stderr, Name ": invalid bitmap magic 0x%x, the bitmap file appears to be corrupted\n", sb->magic);
+		pr_err("invalid bitmap magic 0x%x, the bitmap file appears to be corrupted\n", sb->magic);
 	}
 	printf("         Version : %d\n", sb->version);
 	if (sb->version < BITMAP_MAJOR_LO ||
 	    sb->version > BITMAP_MAJOR_HI) {
-		fprintf(stderr, Name ": unknown bitmap version %d, either the bitmap file is corrupted or you need to upgrade your tools\n", sb->version);
+		pr_err("unknown bitmap version %d, either the bitmap file is corrupted or you need to upgrade your tools\n", sb->version);
 		goto free_info;
 	}
 
 	rv = 0;
-	if (st && st->ss->swapuuid) {
-	printf("            UUID : %08x.%08x.%08x.%08x\n",
-					swapl(*(__u32 *)(sb->uuid+0)),
-					swapl(*(__u32 *)(sb->uuid+4)),
-					swapl(*(__u32 *)(sb->uuid+8)),
-					swapl(*(__u32 *)(sb->uuid+12)));
-	} else {
-	printf("            UUID : %08x.%08x.%08x.%08x\n",
-					*(__u32 *)(sb->uuid+0),
-					*(__u32 *)(sb->uuid+4),
-					*(__u32 *)(sb->uuid+8),
-					*(__u32 *)(sb->uuid+12));
-	}
+	if (st)
+		swap = st->ss->swapuuid;
+	else
+#if __BYTE_ORDER == BIG_ENDIAN
+		swap = 0;
+#else
+		swap = 1;
+#endif
+	memcpy(uuid32, sb->uuid, 16);
+	if (swap)
+		printf("            UUID : %08x:%08x:%08x:%08x\n",
+		       swapl(uuid32[0]),
+		       swapl(uuid32[1]),
+		       swapl(uuid32[2]),
+		       swapl(uuid32[3]));
+	else
+		printf("            UUID : %08x:%08x:%08x:%08x\n",
+		       uuid32[0],
+		       uuid32[1],
+		       uuid32[2],
+		       uuid32[3]);
+
 	printf("          Events : %llu\n", (unsigned long long)sb->events);
 	printf("  Events Cleared : %llu\n", (unsigned long long)sb->events_cleared);
 	printf("           State : %s\n", bitmap_state(sb->state));
@@ -308,7 +321,7 @@ int ExamineBitmap(char *filename, int brief, struct supertype *st)
 		goto free_info;
 	printf("          Bitmap : %llu bits (chunks), %llu dirty (%2.1f%%)\n",
 			info->total_bits, info->dirty_bits,
-			100.0 * info->dirty_bits / (info->total_bits + 1));
+			100.0 * info->dirty_bits / (info->total_bits?:1));
 free_info:
 	free(info);
 	return rv;
@@ -331,13 +344,13 @@ int CreateBitmap(char *filename, int force, char uuid[16],
 	long long bytes, filesize;
 
 	if (!force && access(filename, F_OK) == 0) {
-		fprintf(stderr, Name ": bitmap file %s already exists, use --force to overwrite\n", filename);
+		pr_err("bitmap file %s already exists, use --force to overwrite\n", filename);
 		return rv;
 	}
 
 	fp = fopen(filename, "w");
 	if (fp == NULL) {
-		fprintf(stderr, Name ": failed to open bitmap file %s: %s\n",
+		pr_err("failed to open bitmap file %s: %s\n",
 			filename, strerror(errno));
 		return rv;
 	}
@@ -345,11 +358,12 @@ int CreateBitmap(char *filename, int force, char uuid[16],
 	if (chunksize == UnSet) {
 		/* We don't want more than 2^21 chunks, as 2^11 fill up one
 		 * 4K page (2 bytes per chunk), and 2^10 address of those
-		 * fill up a 4K indexing page.  2^20 might be safer...
+		 * fill up a 4K indexing page.  2^20 might be safer, especially
+		 * on 64bit hosts, so use that.
 		 */
 		chunksize = DEFAULT_BITMAP_CHUNK;
-		/* <<21 for 2^21 chunks, >>9 to convert bytes to sectors */
-		while (array_size > (chunksize << (21-9)))
+		/* <<20 for 2^20 chunks, >>9 to convert bytes to sectors */
+		while (array_size > ((unsigned long long)chunksize << (20-9)))
 			chunksize <<= 1;
 	}
 
@@ -366,7 +380,7 @@ int CreateBitmap(char *filename, int force, char uuid[16],
 	sb_cpu_to_le(&sb); /* convert to on-disk byte ordering */
 
 	if (fwrite(&sb, sizeof(sb), 1, fp) != 1) {
-		fprintf(stderr, Name ": failed to write superblock to bitmap file %s: %s\n", filename, strerror(errno));
+		pr_err("failed to write superblock to bitmap file %s: %s\n", filename, strerror(errno));
 		goto out;
 	}
 
@@ -383,13 +397,14 @@ int CreateBitmap(char *filename, int force, char uuid[16],
 
 	while (bytes > 0) {
 		if (fwrite(block, sizeof(block), 1, fp) != 1) {
-			fprintf(stderr, Name ": failed to write bitmap file %s: %s\n", filename, strerror(errno));
+			pr_err("failed to write bitmap file %s: %s\n", filename, strerror(errno));
 			goto out;
 		}
 		bytes -= sizeof(block);
 	}
-	
+
 	rv = 0;
+	fflush(fp);
 	/* make the file be the right size (well, to the nearest byte) */
 	if (ftruncate(fileno(fp), filesize))
 		perror("ftrunace");
@@ -400,7 +415,7 @@ out:
 	return rv;
 }
 
-int bitmap_update_uuid(int fd, int *uuid)
+int bitmap_update_uuid(int fd, int *uuid, int swap)
 {
 	struct bitmap_super_s bm;
 	if (lseek(fd, 0, 0) != 0)
@@ -409,7 +424,7 @@ int bitmap_update_uuid(int fd, int *uuid)
 		return 1;
 	if (bm.magic != __cpu_to_le32(BITMAP_MAGIC))
 		return 1;
-	memcpy(bm.uuid, uuid, 16);
+	copy_uuid(bm.uuid, uuid, swap);
 	if (lseek(fd, 0, 0) != 0)
 		return 2;
 	if (write(fd, &bm, sizeof(bm)) != sizeof(bm)) {
